"""Residual Neural Networks (ResNets).

This implementation largely follows
[``torchvision.models.resnet``](https://github.com/pytorch/vision/blob/95f10a4ec9e43b2c8072ae5a68edd5700f9b1e45/torchvision/models/resnet.py).
"""

from __future__ import annotations

from collections.abc import Sequence
import copy
from typing import TYPE_CHECKING, Callable, Literal

import torch
from torch import nn
import torchvision

from .. import bnn
from ..bnn import params

if TYPE_CHECKING:
    from jaxtyping import Float
    from torch import Tensor


class ResNet(bnn.BNNModule):
    """A residual neural network for image classification.

    :param out_size: Size of the output (i.e. number of classes).
    :param block: Block type to use.
    :param num_blocks_per_layer: Number of blocks per layer.
    :param zero_init_residual:
    :param groups: Number of groups for the convolutional layers.
    :param width_per_group: Width per group for the convolutional layers.
    :param replace_stride_with_dilation: Whether to replace the 2x2 stride with a dilated
        convolution. Must be a tuple of length 3.
    :param norm_layer: Normalization layer to use.
    :param architecture: Type of ResNet architecture. Either "imagenet" or "cifar".
    :param parametrization: The parametrization to use. Defines the initialization
        and learning rate scaling for the parameters of the module.
    :param cov: Covariance structure of the probabilistic layers.
    """

    def __init__(
        self,
        out_size: int,
        block: type["BasicBlock"] | type["Bottleneck"],
        num_blocks_per_layer: Sequence[int],
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation: Sequence[bool] = (False, False, False),
        norm_layer: Callable[..., nn.Module] = lambda c: nn.GroupNorm(
            num_groups=32, num_channels=c
        ),
        architecture: Literal["imagenet", "cifar"] = "imagenet",
        parametrization: params.Parametrization = params.MaximalUpdate(),
        cov: params.FactorizedCovariance | None = None,
    ) -> None:

        super().__init__(parametrization=parametrization)

        # Attributes
        if norm_layer is nn.BatchNorm2d:
            raise ValueError(
                "BatchNorm is currently not supported due to incompatibility of "
                "torch.vmap with the 'running_stats' tracked by BatchNorm."
                "See also: https://pytorch.org/docs/stable/func.batch_norm.html#patching-batch-norm."
            )
        self._norm_layer = norm_layer
        self.inplanes = 64
        self.dilation = 1
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "'replace_stride_with_dilation' should be "
                f"a 3-element tuple or list, got {replace_stride_with_dilation}"
            )
        self.groups = groups
        self.base_width = width_per_group

        # Layers
        if architecture == "cifar":
            self.conv1 = bnn.Conv2d(
                3,
                self.inplanes,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
                cov=(
                    copy.deepcopy(cov)
                    if isinstance(cov, params.DiagonalCovariance)
                    else copy.deepcopy(cov)
                ),
                parametrization=self.parametrization,
                layer_type="input",
            )
        elif architecture == "imagenet":
            self.conv1 = bnn.Conv2d(
                3,
                self.inplanes,
                kernel_size=7,
                stride=2,
                padding=3,
                bias=False,
                cov=(
                    copy.deepcopy(cov)
                    if isinstance(cov, params.DiagonalCovariance)
                    else copy.deepcopy(cov)
                ),
                parametrization=self.parametrization,
                layer_type="input",
            )
        else:
            raise ValueError(
                f"Unknown architecture '{architecture}'. Expected 'cifar' or 'imagenet'."
            )
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        if architecture == "imagenet":
            self.optional_pool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        else:
            self.optional_pool = None
        self.layer1 = self._make_layer(
            block,
            64,
            num_blocks_per_layer[0],
            parametrization=parametrization,
            cov=(
                copy.deepcopy(cov)
                if isinstance(cov, params.DiagonalCovariance)
                else None
            ),
            layer_type="hidden",
        )
        self.layer2 = self._make_layer(
            block,
            128,
            num_blocks_per_layer[1],
            stride=2,
            dilate=replace_stride_with_dilation[0],
            parametrization=parametrization,
            cov=(
                copy.deepcopy(cov)
                if isinstance(cov, params.DiagonalCovariance)
                else None
            ),
            layer_type="hidden",
        )
        self.layer3 = self._make_layer(
            block,
            256,
            num_blocks_per_layer[2],
            stride=2,
            dilate=replace_stride_with_dilation[1],
            parametrization=parametrization,
            cov=(
                copy.deepcopy(cov)
                if isinstance(cov, params.DiagonalCovariance)
                else None
            ),
            layer_type="hidden",
        )
        self.layer4 = self._make_layer(
            block,
            512,
            num_blocks_per_layer[3],
            stride=2,
            dilate=replace_stride_with_dilation[2],
            parametrization=parametrization,
            cov=(
                copy.deepcopy(cov)
                if isinstance(cov, params.DiagonalCovariance)
                else None
            ),
            layer_type="hidden",
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = bnn.Linear(
            512 * block.expansion,
            out_size,
            parametrization=parametrization,
            cov=(
                copy.deepcopy(cov)
                if isinstance(cov, params.DiagonalCovariance)
                else copy.deepcopy(cov)
            ),
            layer_type="output",
        )

        # Reset parameters
        self.reset_parameters()

        # Zero-initialize the last bottleneck in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck) and m.bn3.weight is not None:
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock) and m.bn2.weight is not None:
                    nn.init.constant_(m.bn2.weight, 0)

    @classmethod
    def from_pretrained_weights(
        cls,
        out_size: int,
        weights: torchvision.models.Weights,
        freeze: bool = False,
        architecture: Literal["imagenet", "cifar"] = "imagenet",
        *args,
        **kwargs,
    ):
        """Load a ResNet model with pretrained weights.

        Depending on the ``out_size`` and ``architecture`` parameters, the first and last
        layers of the model are not initialized with the pretrained weights.

        :param out_size: Size of the output (i.e. number of classes).
        :param weights: Pretrained weights to use.
        :param freeze: Whether to freeze the pretrained weights.
        :param architecture: Type of ResNet architecture. Either "imagenet" or "cifar".
        """
        # Load and preprocess the pretrained weights
        pretrained_weights = weights.get_state_dict(progress=True)
        if architecture != "imagenet":
            # Remove the first layer (conv1) from the pretrained weights
            del pretrained_weights["conv1.weight"]

        if out_size != pretrained_weights["fc.weight"].shape[0]:
            # Remove the last layer (fc) from the pretrained weights
            del pretrained_weights["fc.weight"]
            del pretrained_weights["fc.bias"]

        # Model
        model = cls(*args, **kwargs, out_size=out_size, architecture=architecture)
        missing_keys, unexpected_keys = model.load_state_dict(
            pretrained_weights, strict=False
        )

        if freeze:
            # Freeze the pretrained weights
            for name, param in model.named_parameters():
                if name.replace(".params", "") in pretrained_weights:
                    # TODO: Should the GroupNorm weight and bias be frozen? ResNet pretrained weights use BatchNorm
                    param.requires_grad = False

        return model

    def _make_layer(
        self,
        block: type["BasicBlock"] | type["Bottleneck"],
        planes: int,
        blocks: int,
        stride: int = 1,
        dilate: bool = False,
        parametrization: params.Parametrization = params.MaximalUpdate(),
        cov: params.FactorizedCovariance | None = None,
        layer_type: Literal["input", "hidden", "output"] = "hidden",
    ) -> nn.Sequential:
        """Make a ResNet layer for a given block type."""
        norm_layer = self._norm_layer
        previous_dilation = self.dilation
        downsample = None
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = bnn.Sequential(
                bnn.Conv2d(
                    self.inplanes,
                    planes * block.expansion,
                    stride=stride,
                    kernel_size=1,
                    bias=False,
                    cov=(
                        copy.deepcopy(cov)
                        if isinstance(cov, params.DiagonalCovariance)
                        else copy.deepcopy(cov)
                    ),
                    parametrization=parametrization,
                    layer_type=layer_type,
                ),
                norm_layer(planes * block.expansion),
                parametrization=parametrization,
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride=stride,
                downsample=downsample,
                groups=self.groups,
                base_width=self.base_width,
                dilation=previous_dilation,
                norm_layer=norm_layer,
                parametrization=parametrization,
                cov=(
                    copy.deepcopy(cov)
                    if isinstance(cov, params.DiagonalCovariance)
                    else copy.deepcopy(cov)
                ),
                layer_type=layer_type,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                    parametrization=parametrization,
                    cov=(
                        copy.deepcopy(cov)
                        if isinstance(cov, params.DiagonalCovariance)
                        else copy.deepcopy(cov)
                    ),
                    layer_type=layer_type,
                )
            )

        return bnn.Sequential(*layers, parametrization=parametrization)

    def _forward_impl(
        self,
        input: Float[Tensor, "*sample batch *in_feature"],
        /,
        sample_shape: torch.Size = torch.Size([]),
        generator: torch.Generator | None = None,
        input_contains_samples: bool = False,
        parameter_samples: dict[str, Float[Tensor, "*sample parameter"]] | None = None,
    ) -> Float[Tensor, "*sample *batch *out_feature"]:
        # See note [TorchScript super()]

        out = self.conv1(
            input,
            sample_shape=sample_shape,
            generator=generator,
            input_contains_samples=input_contains_samples,
            parameter_samples=parameter_samples,
        )
        out = bnn.batched_forward(self.bn1, num_batch_dims=len(sample_shape) + 1)(out)
        out = self.relu(out)
        if self.optional_pool is not None:
            out = bnn.batched_forward(
                self.optional_pool, num_batch_dims=len(sample_shape) + 1
            )(out)

        out = self.layer1(
            out,
            sample_shape=sample_shape,
            generator=generator,
            input_contains_samples=True,
            parameter_samples=parameter_samples,
        )
        out = self.layer2(
            out,
            sample_shape=sample_shape,
            generator=generator,
            input_contains_samples=True,
            parameter_samples=parameter_samples,
        )
        out = self.layer3(
            out,
            sample_shape=sample_shape,
            generator=generator,
            input_contains_samples=True,
            parameter_samples=parameter_samples,
        )
        out = self.layer4(
            out,
            sample_shape=sample_shape,
            generator=generator,
            input_contains_samples=True,
            parameter_samples=parameter_samples,
        )

        out = bnn.batched_forward(self.avgpool, num_batch_dims=len(sample_shape) + 1)(
            out
        )
        out = torch.flatten(out, -3)

        final_layer_forward_kwargs = {}
        if isinstance(self.fc, bnn.Linear):
            # This supports swapping out the final layer for a nn.Linear layer
            # e.g. in the case of Laplace approximation.
            final_layer_forward_kwargs = {
                "sample_shape": sample_shape,
                "generator": generator,
                "input_contains_samples": True,
                "parameter_samples": parameter_samples,
            }
        out = self.fc(
            out,
            **final_layer_forward_kwargs,
        )

        return out

    def forward(
        self,
        input: Float[Tensor, "*sample batch *in_feature"],
        /,
        sample_shape: torch.Size = torch.Size([]),
        generator: torch.Generator | None = None,
        input_contains_samples: bool = False,
        parameter_samples: dict[str, Float[Tensor, "*sample parameter"]] | None = None,
    ) -> Float[Tensor, "*sample *batch *out_feature"]:
        return self._forward_impl(
            input,
            sample_shape=sample_shape,
            generator=generator,
            input_contains_samples=input_contains_samples,
            parameter_samples=parameter_samples,
        )


class BasicBlock(bnn.BNNModule):
    """Basic block of a ResNet."""

    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: nn.Module | None = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Callable[..., nn.Module] = nn.BatchNorm2d,
        parametrization: params.Parametrization = params.MaximalUpdate(),
        cov: params.FactorizedCovariance | None = None,
        layer_type: Literal["input", "hidden", "output"] = "hidden",
    ) -> None:
        super().__init__()

        # Attributes
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports 'groups=1' and 'base_width=64'.")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock.")

        # Layers
        self.conv1 = bnn.Conv2d(
            inplanes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=1,
            groups=1,
            dilation=1,
            bias=False,
            parametrization=parametrization,
            cov=(
                copy.deepcopy(cov)
                if isinstance(cov, params.DiagonalCovariance)
                else copy.deepcopy(cov)
            ),
            layer_type=layer_type,
        )
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = bnn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            stride=1,
            padding=1,
            groups=1,
            dilation=1,
            bias=False,
            parametrization=parametrization,
            cov=(
                copy.deepcopy(cov)
                if isinstance(cov, params.DiagonalCovariance)
                else copy.deepcopy(cov)
            ),
            layer_type=layer_type,
        )
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(
        self,
        input: Float[Tensor, "*sample batch *in_feature"],
        /,
        sample_shape: torch.Size = torch.Size([]),
        generator: torch.Generator | None = None,
        input_contains_samples: bool = False,
        parameter_samples: dict[str, Float[Tensor, "*sample parameter"]] | None = None,
    ) -> Float[Tensor, "*sample *batch *out_feature"]:
        identity = input

        out = self.conv1(
            input,
            sample_shape=sample_shape,
            generator=generator,
            input_contains_samples=input_contains_samples,
            parameter_samples=parameter_samples,
        )
        out = bnn.batched_forward(self.bn1, num_batch_dims=len(sample_shape) + 1)(out)
        out = self.relu(out)

        out = self.conv2(
            out,
            sample_shape=sample_shape,
            generator=generator,
            input_contains_samples=True,
            parameter_samples=parameter_samples,
        )
        out = bnn.batched_forward(self.bn2, num_batch_dims=len(sample_shape) + 1)(out)

        if self.downsample is not None:
            identity = self.downsample(
                input,
                sample_shape=sample_shape,
                generator=generator,
                input_contains_samples=True,
                parameter_samples=parameter_samples,
            )

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(bnn.BNNModule):
    """Bottleneck block of a ResNet.

    Compared to the original implementation this places the stride for downsampling at the
    3x3 convolution(self.conv2), rather than at the first 1x1 convolution(self.conv1).
    This variant is also known as ResNet V1.5 and improves accuracy according to
    https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch."
    """

    expansion: int = 4

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: nn.Module | None = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Callable[..., nn.Module] = nn.BatchNorm2d,
        parametrization: params.Parametrization = params.MaximalUpdate(),
        cov: params.FactorizedCovariance | None = None,
        layer_type: Literal["input", "hidden", "output"] = "hidden",
    ) -> None:
        super().__init__()

        # Attributes
        width = int(planes * (base_width / 64.0)) * groups

        # Layers
        self.conv1 = bnn.Conv2d(
            inplanes,
            width,
            stride=1,
            kernel_size=1,
            bias=False,
            parametrization=parametrization,
            cov=(
                copy.deepcopy(cov)
                if isinstance(cov, params.DiagonalCovariance)
                else copy.deepcopy(cov)
            ),
            layer_type=layer_type,
        )
        self.bn1 = norm_layer(width)
        self.conv2 = bnn.Conv2d(
            width,
            width,
            kernel_size=3,
            stride=stride,
            groups=groups,
            dilation=dilation,
            padding=dilation,
            bias=False,
            parametrization=parametrization,
            cov=(
                copy.deepcopy(cov)
                if isinstance(cov, params.DiagonalCovariance)
                else copy.deepcopy(cov)
            ),
            layer_type=layer_type,
        )
        self.bn2 = norm_layer(width)
        self.conv3 = bnn.Conv2d(
            width,
            planes * self.expansion,
            stride=1,
            kernel_size=1,
            bias=False,
            parametrization=parametrization,
            cov=(
                copy.deepcopy(cov)
                if isinstance(cov, params.DiagonalCovariance)
                else copy.deepcopy(cov)
            ),
            layer_type=layer_type,
        )
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(
        self,
        input: Float[Tensor, "*sample batch *in_feature"],
        /,
        sample_shape: torch.Size = torch.Size([]),
        generator: torch.Generator | None = None,
        input_contains_samples: bool = False,
        parameter_samples: dict[str, Float[Tensor, "*sample parameter"]] | None = None,
    ) -> Float[Tensor, "*sample *batch *out_feature"]:
        identity = input

        out = self.conv1(
            input,
            sample_shape=sample_shape,
            generator=generator,
            input_contains_samples=input_contains_samples,
            parameter_samples=parameter_samples,
        )
        out = bnn.batched_forward(self.bn1, num_batch_dims=len(sample_shape) + 1)(out)
        out = self.relu(out)

        out = self.conv2(
            out,
            sample_shape=sample_shape,
            generator=generator,
            input_contains_samples=True,
            parameter_samples=parameter_samples,
        )
        out = bnn.batched_forward(self.bn2, num_batch_dims=len(sample_shape) + 1)(out)
        out = self.relu(out)

        out = self.conv3(
            out,
            sample_shape=sample_shape,
            generator=generator,
            input_contains_samples=True,
            parameter_samples=parameter_samples,
        )
        out = bnn.batched_forward(self.bn3, num_batch_dims=len(sample_shape) + 1)(out)

        if self.downsample is not None:
            identity = self.downsample(
                input,
                sample_shape=sample_shape,
                generator=generator,
                input_contains_samples=True,
                parameter_samples=parameter_samples,
            )

        out += identity
        out = self.relu(out)

        return out


class ResNet18(ResNet):
    """ResNet-18

    :param **kwargs: Additional keyword arguments passed on to :class:`~inferno.bnn.models.ResNet`.
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            *args, block=BasicBlock, num_blocks_per_layer=[2, 2, 2, 2], **kwargs
        )

    @classmethod
    def from_pretrained_weights(
        cls,
        out_size: int,
        architecture: Literal["imagenet", "cifar"] = "imagenet",
        weights: torchvision.models.Weights = torchvision.models.ResNet18_Weights.DEFAULT,
        freeze: bool = False,
        *args,
        **kwargs,
    ):
        return super().from_pretrained_weights(
            out_size=out_size,
            architecture=architecture,
            weights=weights,
            freeze=freeze,
            *args,
            **kwargs,
        )


class ResNet34(ResNet):
    """ResNet-34

    :param **kwargs: Additional keyword arguments passed on to :class:`~inferno.bnn.models.ResNet`.
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            *args, block=BasicBlock, num_blocks_per_layer=[3, 4, 6, 3], **kwargs
        )

    @classmethod
    def from_pretrained_weights(
        cls,
        out_size: int,
        architecture: Literal["imagenet", "cifar"] = "imagenet",
        weights: torchvision.models.Weights = torchvision.models.ResNet34_Weights.DEFAULT,
        freeze: bool = False,
        *args,
        **kwargs,
    ):
        return super().from_pretrained_weights(
            out_size=out_size,
            architecture=architecture,
            weights=weights,
            freeze=freeze,
            *args,
            **kwargs,
        )


class ResNet50(ResNet):
    """ResNet-50

    :param **kwargs: Additional keyword arguments passed on to :class:`~inferno.bnn.models.ResNet`.
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            *args, block=Bottleneck, num_blocks_per_layer=[3, 4, 6, 3], **kwargs
        )

    @classmethod
    def from_pretrained_weights(
        cls,
        out_size: int,
        architecture: Literal["imagenet", "cifar"] = "imagenet",
        weights: torchvision.models.Weights = torchvision.models.ResNet50_Weights.DEFAULT,
        freeze: bool = False,
        *args,
        **kwargs,
    ):
        return super().from_pretrained_weights(
            out_size=out_size,
            architecture=architecture,
            weights=weights,
            freeze=freeze,
            *args,
            **kwargs,
        )


class ResNet101(ResNet):
    """ResNet-101

    :param **kwargs: Additional keyword arguments passed on to :class:`~inferno.bnn.models.ResNet`.
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            *args, block=Bottleneck, num_blocks_per_layer=[3, 4, 23, 3], **kwargs
        )

    @classmethod
    def from_pretrained_weights(
        cls,
        out_size: int,
        architecture: Literal["imagenet", "cifar"] = "imagenet",
        weights: torchvision.models.Weights = torchvision.models.ResNet101_Weights.DEFAULT,
        freeze: bool = False,
        *args,
        **kwargs,
    ):
        return super().from_pretrained_weights(
            out_size=out_size,
            architecture=architecture,
            weights=weights,
            freeze=freeze,
            *args,
            **kwargs,
        )


class ResNeXt50_32X4D(ResNet):
    """ResNext-50 (32x4d)

    :param **kwargs: Additional keyword arguments passed on to :class:`~inferno.bnn.models.ResNet`.
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            *args,
            block=Bottleneck,
            num_blocks_per_layer=[3, 4, 6, 3],
            groups=32,
            width_per_group=4,
            **kwargs,
        )

    @classmethod
    def from_pretrained_weights(
        cls,
        out_size: int,
        architecture: Literal["imagenet", "cifar"] = "imagenet",
        weights: torchvision.models.Weights = torchvision.models.ResNeXt50_32X4D_Weights.DEFAULT,
        freeze: bool = False,
        *args,
        **kwargs,
    ):
        return super().from_pretrained_weights(
            out_size=out_size,
            architecture=architecture,
            weights=weights,
            freeze=freeze,
            *args,
            **kwargs,
        )


class ResNeXt101_32X8D(ResNet):
    """ResNext-101 (32x8d)

    :param **kwargs: Additional keyword arguments passed on to :class:`~inferno.bnn.models.ResNet`.
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            *args,
            block=Bottleneck,
            num_blocks_per_layer=[3, 4, 23, 3],
            groups=32,
            width_per_group=8,
            **kwargs,
        )

    @classmethod
    def from_pretrained_weights(
        cls,
        out_size: int,
        architecture: Literal["imagenet", "cifar"] = "imagenet",
        weights: torchvision.models.Weights = torchvision.models.ResNeXt101_32X8D_Weights.DEFAULT,
        freeze: bool = False,
        *args,
        **kwargs,
    ):
        return super().from_pretrained_weights(
            out_size=out_size,
            architecture=architecture,
            weights=weights,
            freeze=freeze,
            *args,
            **kwargs,
        )


class ResNeXt101_64X4D(ResNet):
    """ResNext-101 (32x4d)

    :param **kwargs: Additional keyword arguments passed on to :class:`~inferno.bnn.models.ResNet`.
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            *args,
            block=Bottleneck,
            num_blocks_per_layer=[3, 4, 23, 3],
            groups=64,
            width_per_group=4,
            **kwargs,
        )

    @classmethod
    def from_pretrained_weights(
        cls,
        out_size: int,
        architecture: Literal["imagenet", "cifar"] = "imagenet",
        weights: torchvision.models.Weights = torchvision.models.ResNeXt101_64X4D_Weights.DEFAULT,
        freeze: bool = False,
        *args,
        **kwargs,
    ):
        return super().from_pretrained_weights(
            out_size=out_size,
            architecture=architecture,
            weights=weights,
            freeze=freeze,
            *args,
            **kwargs,
        )


class WideResNet50(ResNet):
    """WideResNet-50-2

    Architecture described in [Wide Residual Networks](https://arxiv.org/abs/1605.06431). The model is the same
    as a ResNet except for the bottleneck number of channels which is twice larger in every block.

    :param **kwargs: Additional keyword arguments passed on to :class:`~inferno.bnn.models.ResNet`.
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            *args,
            block=Bottleneck,
            num_blocks_per_layer=[3, 4, 6, 3],
            width_per_group=64 * 2,
            **kwargs,
        )

    @classmethod
    def from_pretrained_weights(
        cls,
        out_size: int,
        architecture: Literal["imagenet", "cifar"] = "imagenet",
        weights: torchvision.models.Weights = torchvision.models.Wide_ResNet50_2_Weights.DEFAULT,
        freeze: bool = False,
        *args,
        **kwargs,
    ):
        return super().from_pretrained_weights(
            out_size=out_size,
            architecture=architecture,
            weights=weights,
            freeze=freeze,
            *args,
            **kwargs,
        )


class WideResNet101(ResNet):
    """WideResNet-101-2

    Architecture described in [Wide Residual Networks](https://arxiv.org/abs/1605.06431). The model is the same
    as a ResNet except for the bottleneck number of channels which is twice larger in every block.

    :param **kwargs: Additional keyword arguments passed on to :class:`~inferno.bnn.models.ResNet`.
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            *args,
            block=Bottleneck,
            num_blocks_per_layer=[3, 4, 23, 3],
            width_per_group=64 * 2,
            **kwargs,
        )

    @classmethod
    def from_pretrained_weights(
        cls,
        out_size: int,
        architecture: Literal["imagenet", "cifar"] = "imagenet",
        weights: torchvision.models.Weights = torchvision.models.Wide_ResNet101_2_Weights.DEFAULT,
        freeze: bool = False,
        *args,
        **kwargs,
    ):
        return super().from_pretrained_weights(
            out_size=out_size,
            architecture=architecture,
            weights=weights,
            freeze=freeze,
            *args,
            **kwargs,
        )



import pathlib

import fire
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import torch
import torchmetrics
import tqdm

import inferno


def toy_regression(
    num_train_data: int = 100,
    num_test_data: int = 200,
    batch_size: int = 16,
    num_hidden_layers: int = 3,
    hidden_width: int = 16,
    num_epochs: int = 500,
    lr: float = 1e-2,
    dtype: torch.dtype = torch.float32,
    device: torch.device = torch.device(
        "cuda" if torch.cuda.is_available() else torch.get_default_device()
    ),
    seed: int = 42,
    plot_data: bool = True,
    plot_learning_curves: bool = True,
    plot_prediction: bool = True,
    outdir: str = "plots",
):

    # Setup
    torch.set_default_dtype(dtype)
    torch.set_default_device(device)
    torch.manual_seed(seed)

    # Create output directory
    if plot_data or plot_learning_curves or plot_prediction:
        pathlib.Path(outdir).mkdir(parents=True, exist_ok=True)

    # Generate training and test data
    with torch.no_grad():

        # Latent function
        def f(x):
            return torch.sigmoid(10 * x) - 0.5

        def y(x, noise_scale=0.02):
            return f(x).squeeze() + noise_scale * torch.randn(x.shape[0])

        # Training and test data
        X_train = torch.concatenate(
            [
                0.35 * torch.rand((num_train_data // 3, 1)) - 1.0,
                0.25 * torch.rand((num_train_data // 3, 1)) + 0.25,
                0.25 * torch.rand((num_train_data // 3, 1)) + 0.75,
            ],
            dim=0,
        )

        y_train = y(X_train)

        X_test = 3 * (torch.rand((num_test_data, 1)) - 0.5)
        y_test = y(X_test)

    train_dataset = torch.utils.data.TensorDataset(X_train, y_train)
    test_dataset = torch.utils.data.TensorDataset(X_test, y_test)

    # Dataloader
    train_dataloader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=batch_size,
        shuffle=True,
    )
    test_dataloader = torch.utils.data.DataLoader(
        test_dataset,
        batch_size=None,
        shuffle=False,
    )

    # Plot training and test data
    if plot_data:
        with torch.no_grad():
            fig, axs = plt.subplots(
                nrows=1,
                ncols=1,
                sharex="row",
                sharey="row",
                squeeze=False,
                figsize=(6, 2),
            )

            # Training data
            axs[0, 0].scatter(
                train_dataset[:][0].detach().cpu().numpy(),
                train_dataset[:][1].detach().cpu().numpy(),
                label=f"Training data",
                color="black",
                zorder=-1,
            )

            # Test data
            axs[0, 0].scatter(
                test_dataset[:][0].detach().cpu().numpy(),
                test_dataset[:][1].detach().cpu().numpy(),
                label=f"Test data",
                color="gray",
                alpha=0.1,
                zorder=-2,
            )

            # Axes
            # axs[0, -1].set(ylim=[-2.5, 2.5])

            # Legend
            handles, labels = axs[0, 0].get_legend_handles_labels()
            fig.legend(
                handles=handles,
                labels=labels,
                loc="upper center",
                bbox_to_anchor=(0.5, 0.0),
                fancybox=False,
                shadow=False,
                ncol=3,
                frameon=False,
            )

            fig.tight_layout()
            plt.savefig(
                pathlib.Path(outdir) / "data.svg",
                bbox_inches="tight",
                pad_inches=0.1,
            )
            plt.close(fig)

    # --8<-- [start:model]
    from torch import nn

    from inferno import bnn

    model = bnn.Sequential(
        inferno.models.MLP(
            in_size=1,
            hidden_sizes=[hidden_width] * num_hidden_layers,
            out_size=1,
            activation_layer=nn.SiLU,  # (1)!
            cov=[bnn.params.FactorizedCovariance()]
            + [None] * (num_hidden_layers - 1)
            + [bnn.params.FactorizedCovariance()],
            bias=True,
        ),
        nn.Flatten(-2, -1),
        parametrization=bnn.params.MUP(),
    )
    # --8<-- [end:model]

    results_list = []

    # --8<-- [start:training]
    # Loss function
    loss_fn = nn.MSELoss()

    # Optimizer
    optimizer = torch.optim.SGD(
        params=model.parameters_and_lrs(
            lr=lr, optimizer="SGD"
        ),  # Sets module-specific learning rates
        lr=lr,
        momentum=0.9,
    )

    # Training loop
    for epoch in tqdm.trange(num_epochs):
        model.train()

        for X_batch, y_batch in iter(train_dataloader):
            optimizer.zero_grad()

            X_batch = X_batch.to(device=device)
            y_batch = y_batch.to(device=device)

            y_batch_pred = model(X_batch)  # Uses a single parameter sample

            loss = loss_fn(y_batch_pred, y_batch)

            loss.backward()
            optimizer.step()
            # --8<-- [end:training]

            if torch.isnan(loss):
                raise ValueError("Loss is NaN")

        with torch.no_grad():
            # Model evaluation
            model.eval()

            test_metric_mse = torchmetrics.MeanSquaredError(squared=True, num_outputs=1)

            for X_test, y_test in iter(test_dataloader):

                # Prediction
                y_test_pred = model(X_test, sample_shape=(16,))

                # Metrics
                test_metric_mse.update(
                    (
                        y_test_pred.mean(dim=0)
                        if isinstance(model, bnn.BNNModule)
                        else y_test_pred
                    ),
                    y_test,
                )

            results_list.append(
                {
                    "Dataset": train_dataset.__class__.__name__,
                    "Model": "VarIBO MLP",
                    "Number of Parameters": sum(
                        p.numel() for p in model.parameters() if p.requires_grad
                    ),
                    "Optimizer": optimizer.__class__.__name__,
                    "LR": lr,
                    "Epoch": epoch,
                    "Loss": loss_fn.__class__.__name__,
                    "Train Loss": loss.item(),
                    "Test MSE": test_metric_mse.compute().item(),
                }
            )
    results_df = pd.DataFrame(results_list)

    if plot_learning_curves:
        # Plot learning curves
        with torch.no_grad():
            fig, axs = plt.subplots(
                nrows=1,
                ncols=2,
                figsize=(7, 3),
            )

            # Train Loss
            sns.lineplot(
                data=results_df,
                x="Epoch",
                y="Train Loss",
                hue="Model",
                ax=axs[0],
                alpha=0.8,
                legend=True,
            )

            # Test MSE
            sns.lineplot(
                data=results_df,
                x="Epoch",
                y="Test MSE",
                hue="Model",
                ax=axs[1],
                alpha=0.8,
                legend=False,
            )

            # Axes
            axs[0].set(yscale="log")
            axs[1].set(yscale="log")

            # Legend
            handles, labels = axs[0].get_legend_handles_labels()
            fig.legend(
                handles=handles,
                labels=labels,
                loc="upper center",
                bbox_to_anchor=(0.5, 0.0),
                fancybox=False,
                shadow=False,
                ncol=len(labels),
                frameon=False,
            )
            axs[0].get_legend().remove()

            fig.align_labels()
            plt.tight_layout()

            plt.savefig(
                pathlib.Path(outdir) / "learning_curves.svg",
                bbox_inches="tight",
                pad_inches=0.1,
            )
            plt.close()

    if plot_prediction:
        # Plot predictions
        with torch.no_grad():
            fig, axs = plt.subplots(
                nrows=1,
                ncols=1,
                sharex="row",
                sharey="row",
                figsize=(7, 3),
            )

            # Inputs to predict at
            x_axis_width_factor = 2.0
            xs = torch.linspace(
                torch.min(X_train) * x_axis_width_factor,
                torch.max(X_train) * x_axis_width_factor,
                1000,
                device=device,
                dtype=dtype,
            ).unsqueeze(-1)

            # Latent function
            axs.plot(
                xs.detach().cpu().numpy(),
                f(xs).detach().cpu().numpy(),
                label=f"Latent function",
                color="black",
                linestyle="--",
                alpha=0.3,
            )

            # Training data
            axs.scatter(
                train_dataset[:][0].detach().cpu().numpy(),
                train_dataset[:][1].detach().cpu().numpy(),
                label=f"Training data",
                color="black",
                zorder=-1,
            )

            # Test data
            axs.scatter(
                test_dataset[:][0].detach().cpu().numpy(),
                test_dataset[:][1].detach().cpu().numpy(),
                label=f"Test data",
                color="gray",
                alpha=0.1,
                zorder=-2,
            )

            # Push-forward of posterior distribution
            model.eval()
            fs_pred = model(xs, sample_shape=(10000,))

            axs.plot(
                xs.detach().cpu().numpy(),
                fs_pred.nanmedian(dim=0).values.detach().cpu().numpy(),
                label=f"Prediction",
                color=f"C{0}",
                zorder=5,
            )
            axs.fill_between(
                xs.flatten().detach().cpu().numpy(),
                torch.quantile(fs_pred, q=0.025, dim=0),
                torch.quantile(fs_pred, q=0.975, dim=0),
                alpha=0.4,
                label=f"Uncertainty",
                color=f"C{0}",
            )

            # Samples
            axs.plot(
                xs.detach().cpu().numpy(),
                fs_pred[0:20].squeeze(-1).mT.detach().cpu().numpy(),
                alpha=0.3,
                linewidth=0.5,
                color=f"C{0}",
            )

            # Legend
            handles, labels = axs.get_legend_handles_labels()
            fig.legend(
                handles=handles,
                labels=labels,
                loc="upper center",
                bbox_to_anchor=(0.5, 0.0),
                fancybox=False,
                shadow=False,
                ncol=5,
                frameon=False,
            )

            fig.align_labels()
            plt.tight_layout()

            plt.savefig(
                pathlib.Path(outdir) / "predictions.svg",
                bbox_inches="tight",
                pad_inches=0.1,
            )
            plt.close()


if __name__ == "__main__":
    fire.Fire(toy_regression)



import math

from numpy import testing as npt
import torch
from torch import nn

from inferno import bnn
from inferno.bnn import params

import pytest


@pytest.mark.parametrize(
    "layer_class,kwargs",
    [
        (bnn.Linear, {}),
        (bnn.Conv1d, {"kernel_size": 1}),
        (bnn.Conv2d, {"kernel_size": 1}),
        (bnn.Conv3d, {"kernel_size": 1}),
    ],
)
@pytest.mark.parametrize("layer_type", ["input", "output", "hidden"])
@pytest.mark.parametrize(
    "cov", [params.FactorizedCovariance(), params.LowRankCovariance(500)]
)
def test_parametrization_initializes_parameters_correctly(
    layer_class,
    kwargs,
    layer_type,
    cov,
):
    torch.manual_seed(348)

    layer = layer_class(
        50,
        20,
        bias=True,
        layer_type=layer_type,
        cov=cov,
        parametrization=params.NeuralTangent(),
        **kwargs,
    )

    weight_init_scale = 1
    bias_init_scale = 1
    numel_mean_parameters = sum(
        p.numel() for name, p in layer.named_parameters() if "cov." not in name
    )
    cov_params_init_scale = weight_init_scale / math.sqrt(numel_mean_parameters)

    npt.assert_allclose(
        layer.weight.detach().numpy().std(),
        weight_init_scale,
        atol=1e-1,
        rtol=1e-1,
    )

    npt.assert_allclose(
        layer.bias.detach().numpy().std(),
        bias_init_scale,
        atol=1e-1,
        rtol=1e-1,
    )

    for name, param in layer.params.cov.named_parameters():
        npt.assert_allclose(
            param.detach().numpy().std(),
            cov_params_init_scale,
            atol=1e-1,
            rtol=1e-1,
        )


@pytest.mark.parametrize(
    "layer_class,kwargs",
    [
        (nn.Linear, {}),
        (nn.Conv1d, {"kernel_size": 1}),
        (nn.Conv2d, {"kernel_size": 1}),
        (nn.Conv3d, {"kernel_size": 1}),
    ],
)
def test_parametrization_initializes_torch_module_parameters_correctly(
    layer_class, kwargs
):
    torch.manual_seed(348)

    class MyModel(bnn.BNNModule):

        def __init__(self):
            super().__init__(parametrization=params.NeuralTangent())
            self.layer = layer_class(
                50,
                20,
                bias=True,
                **kwargs,
            )
            self.reset_parameters()

        def forward(self, input):
            return self.layer(input)

    model = MyModel()

    weight_init_scale = 1
    bias_init_scale = 1

    npt.assert_allclose(
        model.layer.weight.detach().numpy().std(),
        weight_init_scale,
        atol=1e-1,
        rtol=1e-1,
    )

    npt.assert_allclose(
        model.layer.bias.detach().numpy().std(),
        bias_init_scale,
        atol=1e-1,
        rtol=1e-1,
    )



"""Basic building blocks for Bayesian neural networks."""

from . import params
from .modules import Conv1d, Conv2d, Conv3d, Linear, Sequential
from .temperature_scaler import TemperatureScaler

from .modules import BNNModule, batched_forward  # isort:skip

__all__ = [
    "BNNModule",
    "Conv1d",
    "Conv2d",
    "Conv3d",
    "Linear",
    "Sequential",
    "TemperatureScaler",
    "batched_forward",
    "params",
]



import numpy.testing as npt
import torch
from torch import nn
import torchvision

import inferno
from inferno.bnn import params

import pytest


@pytest.mark.parametrize(
    "inferno_mlp,torchvision_mlp",
    [
        (
            inferno.models.MLP(
                in_size=100,
                hidden_sizes=[64, 64],
                out_size=10,
                norm_layer=None,
                activation_layer=nn.ReLU,
                bias=True,
                cov=None,
            ),
            torchvision.ops.MLP(
                in_channels=100,
                hidden_channels=[64, 64, 10],
                norm_layer=None,
                activation_layer=nn.ReLU,
                inplace=None,
                bias=True,
            ),
        ),
        (
            inferno.models.MLP(
                in_size=100,
                hidden_sizes=[64, 64],
                out_size=10,
                norm_layer=nn.LayerNorm,
                activation_layer=nn.SiLU,
                dropout=0.5,
                bias=True,
                cov=None,
            ),
            torchvision.ops.MLP(
                in_channels=100,
                hidden_channels=[64, 64, 10],
                norm_layer=nn.LayerNorm,
                activation_layer=nn.SiLU,
                dropout=0.5,
                inplace=None,
                bias=True,
            ),
        ),
    ],
)
def test_same_as_torchvision_mlp(inferno_mlp, torchvision_mlp):
    """Test whether the implementation matches the one of torchvision if no covariance is used."""
    torch.manual_seed(0)

    # Load weights from torchvision model
    state_dict = torchvision_mlp.state_dict()
    inferno_mlp.load_state_dict(state_dict, strict=True)

    # Create random input
    input = torch.randn((2, 100))

    # Forward pass through both models
    inferno_mlp.eval()
    torchvision_mlp.eval()
    with torch.no_grad():
        inferno_output = inferno_mlp(input, sample_shape=(2,))[
            0
        ]  # Draw multiple samples to check batch compatibility

        torchvision_output = torchvision_mlp(input)

    # Compare the outputs
    npt.assert_allclose(
        inferno_output.detach().numpy(),
        torchvision_output.detach().numpy(),
        rtol=1e-5,
        atol=1e-5,
    )


@pytest.mark.parametrize(
    "model",
    [
        inferno.models.MLP(
            in_size=10,
            hidden_sizes=[32, 32],
            out_size=10,
            norm_layer=nn.LayerNorm,
            activation_layer=nn.SiLU,
            dropout=0.1,
            bias=True,
            cov=params.KroneckerCovariance(),
        ),
        inferno.models.MLP(
            in_size=10,
            hidden_sizes=[32, 32],
            out_size=10,
            norm_layer=nn.LayerNorm,
            activation_layer=nn.SiLU,
            dropout=0.1,
            bias=True,
            cov=[None, None, params.KroneckerCovariance()],
        ),
    ],
)
def test_draw_samples(model):
    """Test whether the model can draw samples."""
    torch.manual_seed(0)

    # Create a ResNet model
    in_size = 10

    # Create random input
    batch_size = 8
    input = torch.randn((batch_size, in_size))

    # Forward pass through the model
    model.eval()
    sample_shape = (10,)
    with torch.no_grad():
        output = model(input, sample_shape=sample_shape)

    # Check the shape of the output
    assert output.shape == (*sample_shape, batch_size, in_size)



from __future__ import annotations

from typing import TYPE_CHECKING

import torch
from torch import nn

if TYPE_CHECKING:
    from jaxtyping import Float
    from torch import Tensor

    from .modules.module import BNNModule


class TemperatureScaler(nn.Module):
    """Temperature scaling.

    Tunes the temperature parameter of the model in the last layer to minimize the negative
    log likelihood of the validation set.

    Based on [On Calibration of Modern Neural Networks](https://arxiv.org/abs/1706.04599).

    :param loss_fn: The loss function to be used for calibration.
    :param lr: Learning rate for the optimizer.
    :param max_iter: Maximum number of iterations per optimization step.
    :param tolerance_grad: Tolerance for the gradient.
    :param tolerance_change: Tolerance for the change in the loss function / parameters.
    :param history_size: Size of the history for the LBFGS optimizer.
    """

    def __init__(
        self,
        loss_fn: nn.Module = nn.CrossEntropyLoss(),
        lr: float = 0.1,
        max_iter: int = 100,
        tolerance_grad: float = 1e-7,
        tolerance_change: float = 1e-9,
        history_size: int = 100,
    ) -> None:
        super().__init__()
        if not isinstance(loss_fn, (nn.BCEWithLogitsLoss, nn.CrossEntropyLoss)):
            raise ValueError(
                f"Loss function must be either BCEWithLogitsLoss or CrossEntropyLoss,"
                f" but is {loss_fn.__class__.__name__}."
            )
        self.lr = lr
        self.loss_fn = loss_fn
        self.max_iter = max_iter
        self.history_size = history_size
        self.tolerance_grad = tolerance_grad
        self.tolerance_change = tolerance_change

    def optimize(
        self,
        model: BNNModule,
        dataloader: torch.utils.data.DataLoader,
    ) -> None:
        """Optimizes the temperature of the model.

        :param model: The model to be calibrated, assumed to return logits.
        :param dataloader: The dataloader for the dataset to calibrate on (typically the validation set).
        """
        # Set the model to evaluation mode to ensure temperature scaling is applied
        model_was_in_training_mode = False
        if model.training:
            model_was_in_training_mode = True
            model.eval()

        # Freeze all parameters except the temperature parameter
        requires_grad_dict = {}
        temperature_parameter = None
        for name, param in model.named_parameters():
            requires_grad_dict[name] = param.requires_grad
            if "temperature" not in name:
                param.requires_grad_(False)
            else:
                param.requires_grad_(True)
                temperature_parameter = param

        if temperature_parameter is None:
            raise ValueError(
                "Model does not have a temperature parameter. "
                "Please ensure the model is a BNNModule with a temperature parameter."
            )

        # Optimizer
        optimizer = torch.optim.LBFGS(
            params=[temperature_parameter],
            lr=self.lr,
            max_iter=self.max_iter,
            tolerance_grad=self.tolerance_grad,
            tolerance_change=self.tolerance_change,
            history_size=self.history_size,
            # line_search_fn="strong_wolfe",
        )

        # Define the closure function for the optimizer
        def closure() -> Tensor:
            optimizer.zero_grad()

            # Get all batches from the validation data
            all_logits = []
            all_targets = []
            for inputs, targets in dataloader:
                logits = model(
                    inputs.to(temperature_parameter.device)
                )  # TODO: how should we pick the number of samples?
                all_logits.append(logits)
                all_targets.append(targets)
            all_logits = torch.cat(all_logits).to(temperature_parameter.device)
            all_targets = torch.cat(all_targets).to(temperature_parameter.device)

            # Compute the loss
            loss = self.loss_fn(all_logits, all_targets)

            loss.backward()
            return loss

        # Optimize temperature
        optimizer.step(closure)

        # Set requires_grad to what it was prior to temperature scaling
        for name, param in model.named_parameters():
            param.requires_grad_(requires_grad_dict[name])
        temperature_parameter.requires_grad_(False)

        if model_was_in_training_mode:
            model.train()



"""Model ensembles.

The implementation largely follows the
[PyTorch documentation on ensembling](https://pytorch.org/tutorials/intermediate/ensembling.html).
"""

from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Iterable

import torch
from torch import func, nn

from ..bnn import params
from ..bnn.modules.module import BNNModule

if TYPE_CHECKING:
    from jaxtyping import Float
    from torch import Tensor


class Ensemble(BNNModule):
    """An ensemble of models.

    This class ensembles multiple models with the same architecture by averaging their predictions.

    :param members: List of models to ensemble.
    """

    def __init__(self, members: Iterable[nn.Module]):

        # Parametrization
        parametrization = None
        for module in members:
            if hasattr(module, "parametrization"):
                # bnn.BNNModule members
                if parametrization is None:
                    parametrization = copy.deepcopy(module.parametrization)
                elif not isinstance(module.parametrization, parametrization.__class__):
                    raise ValueError(
                        "All models in the ensemble must have the same parametrization."
                    )
            else:
                # nn.Module members
                parametrization = params.Standard()

        super().__init__(parametrization=parametrization)

        # Construct a "stateless" version of one of the models. It is "stateless" in
        # the sense that the parameters are meta Tensors and do not have storage.
        self.base_module = [copy.deepcopy(next(iter(members))).to(device="meta")]
        # NOTE: Assigning a list is a hack to avoid registering the base module as a submodule,
        # which can cause errors when using .apply, since it is on the meta device.

        self.members = nn.ModuleList(members)

    def forward(
        self,
        input: Float[Tensor, "*batch in_feature"],
        /,
        sample_shape: torch.Size = torch.Size([]),
        generator: torch.Generator | None = None,
        input_contains_samples: bool = False,
        parameter_samples: dict[str, Float[Tensor, "*sample parameter"]] | None = None,
    ) -> Float[Tensor, "*sample *batch out_feature"]:

        def functional_base_module_call(params, buffers, x):
            kwargs = (
                {
                    "sample_shape": sample_shape,
                    "generator": generator,
                    "input_contains_samples": input_contains_samples,
                    "parameter_samples": parameter_samples,
                }
                if isinstance(self.base_module[0], BNNModule)
                else {}
            )
            return func.functional_call(
                self.base_module[0], (params, buffers), (x,), kwargs=kwargs
            )

        # Combine the states of each module by stacking parameters and buffers together.
        # NOTE: These are new leaf tensors in the graph, so they are independent of the member parameters!
        # In particular this means you cannot train the ensemble by using forward.
        params, buffers = func.stack_module_state(self.members)

        # Predict with each member of the ensemble in vectorized form.
        ensemble_forward_eval = torch.vmap(
            functional_base_module_call, in_dims=(0, 0, None), randomness="different"
        )(params, buffers, input)

        return ensemble_forward_eval



from __future__ import annotations

import copy
from typing import Callable

from torch import nn

from inferno import bnn
from inferno.bnn import params


class MLP(bnn.Sequential):
    """A fully-connected feedforward neural network with the same activation function in each layer.

    :param in_size:             Size of the input.
    :param hidden_sizes:        List of hidden layer sizes.
    :param out_size:            Size of the output (e.g. number of classes).
    :param norm_layer:          Normalization layer which will be stacked on top of the linear layer.
    :param activation_layer:    Activation function following a linear layer.
    :param inplace:             Whether to apply the activation function and dropout inplace.
            Default is ``None``, which uses the respective default values.
    :param bias:                Whether to use bias in the linear layer.``
    :param dropout:             The probability for the dropout layer.
    :param parametrization:     The parametrization to use. Defines the initialization
        and learning rate scaling for the parameters of the module.
    :param cov:                 Covariance structure of the weights.
    """

    def __init__(
        self,
        in_size: int,
        hidden_sizes: list[int],
        out_size: int,
        norm_layer: Callable[..., nn.Module] | None = None,
        activation_layer: Callable[..., nn.Module] | None = nn.ReLU,
        inplace: bool | None = None,
        bias: bool = True,
        dropout: float = 0.0,
        parametrization: params.Parametrization = params.MaximalUpdate(),
        cov: (
            params.FactorizedCovariance | list[params.FactorizedCovariance] | None
        ) = None,
    ) -> None:

        self.in_size = in_size
        self.hidden_sizes = hidden_sizes
        self.out_size = out_size
        params = {} if inplace is None else {"inplace": inplace}
        cov_list = []
        if isinstance(cov, list):
            cov_list = cov
            if len(cov_list) != len(hidden_sizes) + 1:
                raise ValueError(
                    "The length of the covariance list must be equal to the number of layers."
                )
        elif isinstance(cov, bnn.params.FactorizedCovariance):
            cov_list = [copy.deepcopy(cov) for _ in range(len(hidden_sizes) + 1)]
        else:
            cov_list = [None for _ in range(len(hidden_sizes) + 1)]

        # Layers
        layers = [
            bnn.Linear(
                in_features=self.in_size,
                out_features=self.hidden_sizes[0],
                bias=bias,
                layer_type="input",
                cov=cov_list[0],
            ),
        ]
        if norm_layer is not None:
            layers.append(norm_layer(self.hidden_sizes[0]))

        if activation_layer is not None:
            layers.append(activation_layer(**params))

        layers.append(nn.Dropout(dropout, **params))

        for idx_layer, (in_features, out_features) in enumerate(
            zip(self.hidden_sizes[0:-1], self.hidden_sizes[1:])
        ):
            layers.append(
                bnn.Linear(
                    in_features=in_features,
                    out_features=out_features,
                    bias=bias,
                    cov=cov_list[idx_layer + 1],
                )
            )
            if norm_layer is not None:
                layers.append(norm_layer(out_features))
            if activation_layer is not None:
                layers.append(activation_layer(**params))

            layers.append(nn.Dropout(dropout, **params))

        layers.append(
            bnn.Linear(
                in_features=(
                    self.hidden_sizes[-1]
                    if len(self.hidden_sizes) > 0
                    else self.in_size
                ),
                out_features=self.out_size,
                bias=bias,
                layer_type="output",
                cov=cov_list[-1],
            )
        )
        layers.append(nn.Dropout(dropout, **params))

        super().__init__(*layers, parametrization=parametrization)



from __future__ import annotations

import abc
import math
from typing import TYPE_CHECKING, Callable, Literal

import torch
from torch import nn

from ..params import MaximalUpdate, Parametrization

if TYPE_CHECKING:
    from jaxtyping import Float
    from torch import Tensor


class BNNModule(nn.Module, abc.ABC):
    """Base class for all Bayesian neural network modules.

    :param parametrization: The parametrization to use. Defines the initialization
        and learning rate scaling for the parameters of the module.
    """

    def __init__(
        self, parametrization: Parametrization | None = MaximalUpdate()
    ) -> None:
        super().__init__()
        self._parametrization = parametrization

    @property
    def parametrization(self) -> Parametrization:
        """Parametrization of the module."""
        return self._parametrization

    @parametrization.setter
    def parametrization(self, new_parametrization) -> Parametrization:
        self._parametrization = new_parametrization

        # Set the parametrization of all children to the new parametrization.
        for layer in self.children():
            if isinstance(layer, BNNModule):
                layer.parametrization = self.parametrization

    def reset_parameters(self) -> None:
        """Reset the parameters of the module and set the parametrization of all children
        to the parametrization of the module.

        This method should be implemented by subclasses to reset the parameters of the module.
        """
        for layer in self.children():
            if isinstance(layer, BNNModule):
                # Set the parametrization of all children to the parametrization of the parent module.
                layer.parametrization = self.parametrization
                # Initialize the parameters of the child module.
                layer.reset_parameters()
            else:
                if hasattr(layer, "reset_parameters"):
                    reset_parameters_of_torch_module(
                        layer, parametrization=self.parametrization
                    )

    def parameters_and_lrs(
        self,
        lr: float,
        optimizer: Literal["SGD", "Adam"],
    ) -> list[dict[str, Tensor | float]]:
        """Get the parameters of the module and their learning rates for the chosen optimizer
        and the parametrization of the module.

        :param lr: The global learning rate.
        :param optimizer: The optimizer being used.
        """
        param_groups = []

        # Cycle through all children of the module and get their parameters and learning rates
        for layer in self.children():

            # For layers with leaf parameters, return them with adjusted learning rate based on
            # the parametrization.
            if isinstance(layer, BNNModule):
                param_groups += layer.parameters_and_lrs(lr=lr, optimizer=optimizer)

            else:
                if len(list(layer.parameters())) > 0:
                    param_groups += parameters_and_lrs_of_torch_module(
                        layer,
                        lr=lr,
                        parametrization=self.parametrization,
                        optimizer=optimizer,
                    )

        return param_groups

    def forward(
        self,
        input: Float[Tensor, "*sample batch *in_feature"],
        /,
        sample_shape: torch.Size = torch.Size([]),
        generator: torch.Generator | None = None,
        input_contains_samples: bool = False,
        parameter_samples: dict[str, Float[Tensor, "*sample parameter"]] | None = None,
    ) -> Float[Tensor, "*sample *batch *out_feature"]:
        """Forward pass of the module.

        :param input: Input tensor.
        :param sample_shape: Shape of samples.
        :param generator: Random number generator.
        :param input_contains_samples: Whether the input already contains
            samples. If True, the input is assumed to have ``len(sample_shape)``
            many leading dimensions containing input samples (typically
            outputs from previous layers).
        :param parameter_samples: Dictionary of parameter samples. Used to pass
            sampled parameters to the module. Useful to jointly sample parameters
            of multiple layers.
        """
        raise NotImplementedError


def reset_parameters_of_torch_module(
    module: nn.Module,
    /,
    parametrization: Parametrization,
) -> None:
    """Reset the parameters of a torch.nn.Module according to a given parametrization.

    :param module: The torch.nn.Module to reset the parameters of.
    :param parametrization: The parametrization to use.
    """
    module_parameter_names = [param_name for param_name, _ in module.named_parameters()]
    if len(module_parameter_names) == 0:
        return
    elif isinstance(
        module,
        (nn.LayerNorm, nn.GroupNorm, nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d),
    ):
        # No need to reset parameters of layer norm according to Appendix B.1 of http://arxiv.org/abs/2203.03466
        return
        # nn.init.normal_(
        #     module.weight,
        #     mean=0,
        #     std=parametrization.weight_init_scale(
        #         fan_in=1.0, fan_out=None, layer_type="input"
        #     ),
        # )
        # if module.bias is not None:
        #     nn.init.normal_(
        #         module.bias,
        #         mean=0,
        #         std=parametrization.bias_init_scale(
        #             fan_in=1.0, fan_out=None, layer_type="input"
        #         ),
        #     )
    elif "weight" in module_parameter_names:
        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(module.weight)

        nn.init.normal_(
            module.weight,
            mean=0,
            std=parametrization.weight_init_scale(
                fan_in=fan_in, fan_out=fan_out, layer_type="hidden"
            ),
        )
        if module.bias is not None:
            nn.init.normal_(
                module.bias,
                mean=0,
                std=parametrization.bias_init_scale(
                    fan_in=fan_in, fan_out=fan_out, layer_type="hidden"
                ),
            )
    else:
        raise NotImplementedError(
            f"Cannot reset parameters of module: {module.__class__.__name__} "
            f"according to the {parametrization.__class__.__name__} parametrization."
        )


def parameters_and_lrs_of_torch_module(
    module: nn.Module,
    /,
    lr: float,
    parametrization: Parametrization,
    optimizer: Literal["SGD", "Adam"],
) -> list[dict[str, Tensor | float]]:
    """Get the parameters and their learning rates for the chosen parametrization and optimizer.

    :param module: The torch.nn.Module to get the parameters and learning rates of.
    :param lr: The global learning rate.
    :param parametrization: The parametrization to use.
    :param optimizer: The optimizer being used.
    """
    param_groups = []
    module_parameter_names = [param_name for param_name, _ in module.named_parameters()]
    if len(module_parameter_names) == 0:
        pass
    elif isinstance(
        module,
        (
            nn.LayerNorm,
            nn.GroupNorm,
            nn.BatchNorm1d,
            nn.BatchNorm2d,
            nn.BatchNorm3d,
        ),
    ):
        fan_out = module.weight.shape.numel()
        param_groups + [
            {
                "params": module.weight,
                "lr": lr
                * parametrization.weight_lr_scale(
                    fan_in=1.0,
                    fan_out=fan_out,
                    optimizer=optimizer,
                    layer_type="input",
                ),
            }
        ]

        if module.bias is not None:
            param_groups + [
                {
                    "params": module.bias,
                    "lr": lr
                    * parametrization.bias_lr_scale(
                        fan_in=1.0,
                        fan_out=fan_out,
                        optimizer=optimizer,
                        layer_type="input",
                    ),
                }
            ]
    elif "weight" in module_parameter_names:
        fan_in, fan_out = nn.init._calculate_fan_in_and_fan_out(module.weight)

        param_groups + [
            {
                "params": module.weight,
                "lr": lr
                * parametrization.weight_lr_scale(
                    fan_in=fan_in, fan_out=fan_out, optimizer=optimizer
                ),
            }
        ]

        if module.bias is not None:
            param_groups + [
                {
                    "params": module.bias,
                    "lr": lr
                    * parametrization.bias_lr_scale(
                        fan_in=fan_in, fan_out=fan_out, optimizer=optimizer
                    ),
                }
            ]
    else:
        raise NotImplementedError(
            f"Cannot set learning rates of module: {module.__class__.__name__} "
            f"according to the {parametrization.__class__.__name__} parametrization."
        )

    return param_groups


def batched_forward(obj: nn.Module, num_batch_dims: int) -> Callable[
    [Float[Tensor, "*sample batch *in_feature"]],
    Float[Tensor, "*sample batch *out_feature"],
]:
    """Call a torch.nn.Module on inputs with arbitrary many batch dimensions rather than
    just a single one.

    This is useful to extend the functionality of a torch.nn.Module to work with arbitrary
    many batch dimensions, for example arbitrary many sampling dimensions.

    :param obj: The torch.nn.Module to call.
    :param num_batch_dims: The number of batch dimensions.
    """
    if num_batch_dims < 0:
        raise ValueError(
            f"num_batch_dims must be non-negative, but is {num_batch_dims}."
        )
    if num_batch_dims <= 1:
        return obj.__call__

    def batched_forward_helper(
        input: Float[Tensor, "*sample *batch *in_feature"],
    ) -> Float[Tensor, "*sample *batch *out_feature"]:
        flattened_input = input.flatten(start_dim=0, end_dim=num_batch_dims - 2)
        flattened_output = torch.vmap(obj.__call__, in_dims=0, out_dims=0)(
            flattened_input
        )
        return flattened_output.unflatten(0, input.shape[: num_batch_dims - 1])

    return batched_forward_helper



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, OrderedDict, overload

import torch
from torch import nn

from ...params import BNNParameter, Parametrization
from ..module import BNNModule, batched_forward

if TYPE_CHECKING:
    from jaxtyping import Float
    from torch import Tensor


class Sequential(BNNModule, nn.Sequential):
    """A sequential container for modules.

    Modules will be added to it in the order they are passed in the
    constructor. Alternatively, an ``OrderedDict`` of modules can be
    passed in. The ``forward()`` method of ``Sequential`` accepts any
    input and forwards it to the first module it contains. It then
    "chains" outputs to inputs sequentially for each subsequent module,
    finally returning the output of the last module.

    The value a ``Sequential`` provides over manually calling a sequence
    of modules is that it allows treating the whole container as a
    single module, such that performing a transformation on the
    ``Sequential`` applies to each of the modules it stores (which are
    each a registered submodule of the ``Sequential``)

    :param *args: Any number of modules to add to the container.
    :param parametrization: The parametrization to use. If `None`, the
        parametrization of the modules in the container will be used. If
        a :class:`~inferno.bnn.params.Parametrization` object is passed,
        it will be used for all modules in the container.
    """

    # TODO: sample parameters jointly depending on the passed Covariance object

    @overload
    def __init__(
        self, *args: BNNModule, parametrization: Parametrization | None = None
    ) -> None: ...

    @overload
    def __init__(
        self,
        arg: OrderedDict[str, BNNModule],
        parametrization: Parametrization | None = None,
    ) -> None: ...

    def __init__(self, *args, parametrization: Parametrization | None = None):
        super().__init__(parametrization=parametrization)

        if len(args) == 1 and isinstance(args[0], OrderedDict):
            for key, module in args[0].items():
                self.add_module(key, module)
        else:
            for idx, module in enumerate(args):
                self.add_module(str(idx), module)

        if parametrization is not None:
            # Recursively set parametrization of children to the given parametrization
            self.parametrization = parametrization
            """Parametrization of the module."""

            # Reset parameters of all modules in the container
            self.reset_parameters()

    def forward(
        self,
        input: Float[Tensor, "*batch in_feature"],
        /,
        sample_shape: torch.Size = torch.Size([]),
        generator: torch.Generator | None = None,
        input_contains_samples: bool = False,
        parameter_samples: dict[str, Float[Tensor, "*sample parameter"]] | None = None,
    ) -> Float[Tensor, "*sample *batch out_feature"]:

        # Sequential forward passes through all modules in the container
        for module in self:

            if isinstance(module, BNNModule):

                if any(
                    [
                        isinstance(submodule, BNNParameter)
                        for submodule in module.modules()
                    ]
                ):
                    # If the module contains BNN parameters, we need to sample

                    if not input_contains_samples:

                        # Repeat input for each sample
                        input = input.expand(*sample_shape, *input.shape)
                        input_contains_samples = True

                    # Forward pass
                    input = module(
                        input,
                        sample_shape=sample_shape,
                        generator=generator,
                        input_contains_samples=input_contains_samples,
                        parameter_samples=parameter_samples,
                    )
                else:
                    # No BNN parameters, so we can call the module directly without expanding the input
                    input = module(
                        input,
                        sample_shape=(
                            sample_shape if input_contains_samples else torch.Size([])
                        ),
                        generator=generator,
                        input_contains_samples=input_contains_samples,
                        parameter_samples=parameter_samples,
                    )
            elif isinstance(module, nn.Module):
                if not input_contains_samples:
                    # No sample dimensions have been added yet, so we can call the module directly
                    input = module(input)
                else:
                    # Some torch.Modules allow for only a single batch dimension.
                    # Since we potentially operate on multiple batch dimensions (i.e *samples batch)
                    # we need to account for this to ensure interoperability with torch.Modules.
                    if isinstance(module, nn.Flatten):
                        new_start_dim = module.start_dim + (
                            len(sample_shape) if module.start_dim >= 0 else 0
                        )
                        new_end_dim = module.end_dim + (
                            len(sample_shape) if module.end_dim >= 0 else 0
                        )
                        input = input.flatten(
                            start_dim=new_start_dim, end_dim=new_end_dim
                        )
                    elif isinstance(module, nn.Unflatten):
                        new_dim = module.dim + (
                            len(sample_shape) if module.dim >= 0 else 0
                        )
                        input = input.unflatten(new_dim, module.unflattened_size)
                    elif isinstance(
                        module,
                        (nn.Linear, nn.Identity)
                        + tuple(
                            (  # Activation functions
                                module
                                for _, module in inspect.getmembers(
                                    torch.nn.modules.activation
                                )
                                if inspect.isclass(module)
                                and module.__name__
                                in torch.nn.modules.activation.__all__
                            )
                        ),
                    ):
                        # Modules which allow for arbitrary many batch dimensions
                        input = module(input)
                    else:
                        # Call the module's forward pass in batched mode
                        input = batched_forward(
                            module, num_batch_dims=len(sample_shape) + 1
                        )(input)
            else:
                raise ValueError(
                    f"Sequential contains unsupported module type: {type(module)}"
                )

        if not input_contains_samples:
            # In case there are no BNN layers, simply expand the output to the sample shape.
            input = input.expand(*sample_shape, *input.shape)

        return input
